/*
 * $Id: dbtable.js,v 1.8 2005/08/11 16:10:40 johnh Exp $
 *
 * COPYRIGHT(C) 2003 Coldwave Programs Ltd. ALL RIGHTS RESERVED
 * for an updated copy of this file please go to www.dbajax.com
 *
 * By downloading this software you have agreed to the following terms:
 * This software is available free for evaluation and review only.
 * If you wish to incorporate ANY of the code in this release into
 * a product or production application a source license is required
 * which can be purchased at http://www.dbajax.com/purchase.html
 * This software may not be redistributed whole or in part.
 *
 * THIS SOFTWARE IS PROVIDED BY COLDWAVE PROGRAMS LTD "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL COLDWAVE PROGRAMS LTD OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

function DBTable() {
}

DBTable.prototype.initUnscroll = function(row) {
  var tab = row.parentNode.parentNode;
  var sp;

  for ( sp = tab.parentNode ; sp ; sp = sp.parentNode )
    if ( sp.currentStyle && sp.currentStyle.overflow == "auto" )
      break;

  row.scrollParent = sp || document.body;

  tab.parentOffsetLeft = 0;
  tab.parentOffsetTop = 0;

  for ( var p = tab ; (p = p.offsetParent) && // p.tagName != "BODY" &&
	p.currentStyle.overflow != "auto" ; ) {
    tab.parentOffsetLeft += p.offsetLeft;
    tab.parentOffsetTop += p.offsetTop;
  }

  return this.doUnscroll( row, row.scrollParent.scrollTop, row.scrollParent.scrollLeft );
}

// called when parent scroll position changes
// this works when there is no table border
DBTable.prototype.doUnscroll = function(row,top,left) {
  var tab = row.parentNode.parentNode;
  var hScroll = Math.max( 0, left-tab.offsetLeft-tab.parentOffsetLeft );
  hScroll = hScroll < tab.clientWidth-50 ? hScroll : 0;

  if ( tab.scrolledLeft != hScroll )
    tab.scrolledLeft = hScroll;

  var vScroll = Math.max( 0, top-tab.offsetTop-tab.parentOffsetTop );
  if ( vScroll && !row.style.position )
    row.style.position = "relative";
  return vScroll < tab.clientHeight-50 ? vScroll : 0;
}

DBTable.prototype.headerClick = function(e) {
  var cell = e ? e.target : event.srcElement;
  if ( !cell.tagName.match( /TD|TH/ ) )
    return true;
  var row = cell.parentNode;
  var tab = row.parentNode.parentNode;
  var cells = row.childNodes;
  var hdrs = tab.childNodes[0].childNodes[0].childNodes;

  if ( cells != hdrs )
    return false;

  var i;

  for( i=0 ; i<hdrs.length ; i++ )
    if ( cell == hdrs[i] )
      break;

  if ( cell.style.fontStyle ) {
    for( i=0 ; i<hdrs.length ; i++ )
      cells[i].style.fontStyle = "";
    i = -1;
  }
  else
    cell.style.fontStyle = "italic";

  dbtable.sortTable( tab, i );
  event.cancelBubble = true;
  return false;
}

DBTable.prototype.sortTable = function(tab,c) {

  // create array object of non-fixed rows  
  if ( !tab.sortRows ) {
    tab.sortRows = [];
    for ( var r=1 ; r<tab.rows.length ; r++ )
      (tab.sortRows[r-1] = tab.rows[r]).idx = r-1;
  }

  // column<0 resets list of sorts
  if ( c<0 || !tab.sortRows.length ) {
    for ( var r=0 ; r<tab.sortRows.length ; r++ )
      tab.sortRows[r].style.top = 0;
    tab.sorts = "";
    return;
  }

  // columns to be storted by are cumulative
  tab.sorts = (tab.sorts ? tab.sorts+"," : "")+c;

  // construct the body of the comparison
  // DBTable.prototype.to be used by the array sort
  var cmp = "return ";
  var cs = (""+tab.sorts).split( "," );
  for( var c=0 ; c<cs.length ; c++ )
    cmp += (c?"||":"")+
	(tab.rows[0].cells[cs[c]].type.match( /int|num|float/ ) ? 
	"(cln(b)-cln(a))" : "(a>b?1:a<b?-1:0)")
	.replace( /(a|b)/g, "$1.cells["+cs[c]+"].innerText.toLowerCase()" );

  // sort rows into new order using dynamic comparison function
  var out = tab.sortRows = tab.sortRows.sort( new Function( "a,b", cmp ) );
  var h = tab.rows[1].clientHeight; // height of a row

  // reposition rows according to their new position
  for( var r=0 ; r<out.length ; r++ ) {
    if (1)
      out[r].style.top = (r-out[r].idx)*h; // simpler and safer
    else
      out[r].style.setExpression( "top", // unreliable on IE5.5
      ""+(r-out[r].idx)+"*"+tab.id+".rows[1].clientHeight" );

    out[r].style.position = "relative";
  }
}

function cln(n) {
//log.value += ">"+n+"<";
  return n=='null' ? -0.001 : n.replace(/,/g,'').replace(/\u2212/,'-');
}

DBTable.prototype.updateAlignment = function( inspected, clientWidth, clientHeight ) {
  var alignment = inspected.alignment;
  if ( alignment=="" )
    return;

    if ( alignment & 1 )
      inspected.style.setExpression( "width", 
	"Math.max( 20, document.body.clientWidth-"+
	(clientWidth-inspected.offsetWidth )+")" );
    else
      inspected.style.setExpression( "width", inspected.offsetWidth );

    if ( alignment & 2 )
      inspected.style.setExpression( "height", 
	"Math.max( 20, document.body.clientHeight-"+
	(clientHeight-inspected.offsetHeight )+")" );
    else
      inspected.style.setExpression( "height", inspected.offsetHeight );

    var parent = inspected.parentNode;

    if ( alignment & 4 )
      parent.style.setExpression( "left", 
	"document.body.clientWidth-"+(clientWidth-parent.offsetLeft) );
    else
      parent.style.setExpression( "left", parent.offsetLeft );

    if ( alignment & 8 )
      parent.style.setExpression( "top", 
	"document.body.clientHeight-"+(clientHeight-parent.offsetTop) );
    else
      parent.style.setExpression( "top", parent.offsetTop );

  if ( !alignment )
    inspected.alignment = "";
}

DBTable.prototype.exportTable = function(id,delim) {
  var tab = document.getElementById( id ).getElementsByTagName("TABLE")[0];
  var text = ""; // rows are exported in sorted order if there is one
  var rows = tab.sortRows ? [tab.rows[0]].concat( tab.sortRows ) : tab.rows;

  for( var r=0; r<rows.length ; r++ ) {
    var cols = rows[r].cells;
    for( var c=0 ; c<cols.length ; c++ )
      text += (c?delim||"\t":r?"\n":"")+cols[c].innerText.replace(/\t/g," ");
  }

  window.clipboardData.setData( "Text", text );
  alert( "Data exported\n  to clipboard" );
}

DBTable.prototype.changePage = function(id,change,el) {
  el = el || document.getElementById(id);
  if( !el ) {
    alert( "changePage: could not find element named: "+id );
    return false;
  }

  var body = el.getElementsByTagName("TBODY")[0];
  if ( !body )
    return false;

  var rows = body.getElementsByTagName("TR");

  if ( el.page )
    for ( var r = (el.page-1) * el.pagerows ; 
	r<el.page*el.pagerows && r < rows.length ; r++ )
      rows[r].style.display = "";

  var maxpage = Math.floor((rows.length-1)/el.pagerows+1);
  el.page = (el.page || 0) + change;
  if ( el.page < 1 ) el.page = 1;
  if ( el.page > maxpage ) el.page = maxpage;

  for ( var r = (el.page-1) * el.pagerows ; 
	r<el.page*el.pagerows && r<rows.length ; r++ )
     rows[r].style.display = "block";
}


DBTable.prototype.firstPage = function(id,el) {
  this.changePage(id,-9999,el);
}
DBTable.prototype.prevPage = function(id,el) {
  this.changePage(id,-1,el);
}
DBTable.prototype.nextPage = function(id,el) {
  this.changePage(id,1,el);
}
DBTable.prototype.lastPage = function(id,el) {
  this.changePage(id,9999,el);
}

DBTable.prototype.printTable = function(id,title,header,footer) {
  var el = document.getElementById(id);
  if( !el ) {
    alert( "printTable: could not find element named: "+id );
    return false;
  }

  var hdrs = el.getElementsByTagName("THEAD")[0].childNodes;
  var rows = el.getElementsByTagName("TBODY")[0].childNodes;
  var win = window.open( "", "", "" );
  var doc = win.document;

  doc.write( '<html><head><link rel="stylesheet" href="/dbml/dbtable.css">'+
	(title?'<title>'+title+'</title>':"")+
	'</head><body><table cellspacing="0" class="dbprint"><thead>' );
  var blank = '<tr><td colspan="100" style="border: none;">';

  for ( var r=0 ; r<rows.length ; r++ ) {
    if ( r%el.printrows == 0 ) {
      if ( header )
        doc.write( blank+header+'</td></tr>' );
      for ( var i=0 ; i<hdrs.length ; i++ )
        doc.write( hdrs[i].outerHTML );
    }
    var row = rows[r].outerHTML;
    if ( r%el.printrows == el.printrows-1 || r==rows.length-1 ) {
      if ( footer )
        row = row+(blank+(footer+'</td></tr>')
		.replace( />/, ' style="page-break-after:always;">' ));
      else if ( r!=rows.length-1 )
        row = row.replace( />/, ' style="page-break-after:always;">' );
    }
    doc.write( row );
  }

  doc.write( "</table>" );

}

dbtable = new DBTable();

if ( !ga( document.body, "dbtable_js" ) ) {
  document.body.dbtable_js = 1;
  document.write( '<link rel="stylesheet" href="/dbml/dbtable.css">' );
}
