/*
    gui.js - $Id: gui.js,v 3.15 2007/06/23 18:09:34 johnh Exp $
    ======

  example graphical user interface for Xtab data grouping/toalling
  class contained in file ouput(e).js. Instantiate this class with
  with either an XML string or a URL to fetch a XML document to
  create an Xtab browsable table.

  Copyright(c) John Holdsworth 2001. All rights reserved.
  
  This software is provided for evaluaton purposes only and without
  Warantee implied or expressed. It must not be redistributed.
*/

//**Start Encode**

// debugging
var timevar = (new Date()).getTime();
var t0 = timevar;

function log (msg) {
  var text = document.all["console"];
  if ( text ) {
    text.value += "\n"+msg;
    text.scrollTop = 1000;
  }
}

function since (msg) {
    var mills = -timevar + (timevar = (new Date()).getTime());
    log( msg+": "+(mills/1000)+" ("+(timevar-t0)/1000+")" );
}

//onload = since;

// utility function to for regexp replace
function String.prototype.subst ( restr, to ) {
   return this.replace( new RegExp( restr ), to||"" );
}

// pull in other the other code required
if ( !window.XTab ) {
  document.write( '<script src="/xtab/menu.js"></script>' );

  // hack for development with clear text version
//  if ( (""+top.document.location).match( "http://plan" ) )
    document.write( '<script src="/xtab/output.js"></script>' );
//  else
//    document.write( '<script src="/xtab/outpute.js"'+
//		' language="JScript.Encode"></script>' );
}

// column number from <td> cell
function colno (cell) {
  var cells = cell.parentElement.children;

  for ( var c = 0 ; c < cells.length ; c++ )
    if ( cells[c] == cell )
	return c;
}

// get the column object for a <td> cell
function cellCol (cell) {
  return xtab.columns[display.split( /,/ )[colno(cell)]];
}

// map context menu on right click
function openMenu () { with (window.gui) {
  var event = window.gui.window.event;

  if ( !event.shiftKey ) {
    var over = event.srcElement;
    over.colno = colno( over );
    over.c0 = over.parentElement.c0;
    over.column = xtab.columns[display.split( /,/ )[over.colno]];
    over.node = xtab.data.selectSingleNode( './/*[@c0="'+over.c0+'"]' )
	|| xtab.data.documentElement;
    over.level = over.parentElement._level||0;

    window.colMenu.children[0].innerText = over.column.name;
    window.contextMenu.children[0].innerText = "["+over.column.name+"]";
    window.contextMenu.children[10].innerText = over.c0 > 0 ? "Close" : "Open All";

    return mapMenu( window.contextMenu, over );
  }
} }

// dispatch menu item on click
function itemClicked () { with (window.gui) {
  var event = window.gui.window.event;
  var item = event.srcElement;
  unmapMenu( "contextMenu" );

  if ( item.callback ) {
    if ( typeof item.callback == "string" ) 
	item.callback = new Function( "", item.callback+"(this)" );

    var over = item.over = document.all.contextMenu.parentItem;

    if ( over ) {
      item.colno = over.colno;
      item.column = over.column;
      item.c0 = over.c0;
      item.node = over.node;
      item.level = over.lovel;
    }

    item.callback();
  }

  return event.returnValue = event.button == 2 ? false : true;
} }

// create main context menu
function initMenus () { with (window.gui) {

  document.body.onclick = itemClicked;

  var menu = new Menu( "contextMenu" ); 
  menu.item( "[column]", null, "selected column" );
  menu.item( "Reset Groups", "reGroup", "Reset all groupings" );
  menu.submenu( "Group by", "colMenu", "Add column to last grouping" );
  menu.submenu( "Total by", "colMenu", "Start new grouping" );
  menu.submenu( "Open by", "colMenu", "Start new grouping, opening cells" );
  menu.item( "Open All", "openGroup", "Open all underlying row cells" );
  menu.item( "Undo", "doUndo' accessKey='k", "undo last operation" );
  menu.item();
  menu.submenu( "Sort", "sortMenu", "Add or reset sort" );
  menu.submenu( "Show", "colMenu", "Add column at selected position" );
  menu.item( "Hide", "doHide", "Hide selected column" );
  menu.item( "Edit", "doEdit", "Edit the value an open cell" );
  menu.item( "Find", "doFind", "Search for value in a cell" );
  menu.close( document );

  var smenu = new Menu( "sortMenu" );
  smenu.item( "Reset Sorts", "unSort", "Reset all sorts");
  smenu.item( "Ascending",   "doSort", "Add ascending sort by this column" );
  smenu.item( "Descending",  "doSort", "add descending sort" );
  smenu.close( document );

  return menu;
} }

// test to see if user has MSXML3 installed
function xml3 () {
  if ( confirm( "Please install XMSXML 3.0 and refresh\n"+
		    "Click OK to download"  ) )
    document.location = "http://download.microsoft.com/download/xml/Install/3.10/W98NT42KMe/EN-US/msxml3sp1.EXE";
}

if ( 0 ) {
document.write( '<object id="MSXML3" \
classid="clsid: f5078f32-c551-11d3-89b9-0000f81fe221" \
codebase="msxml3.cab#version=8,10,8308,0" \
type="application/x-oleobject" \
STYLE="display: none"> \
</object>' );
}

document.write( "<link rel='stylesheet' href='/xtab/style.css'>" );

var gno = "";

// create a "Gui" object
function Gui (island, benchmark, aWindow) {
  this.gno = gno; gno++;
  return this.loadup( island, benchmark, aWindow );
}

function Gui.prototype.loadup (island, benchmark, aWindow) {
  this.window = aWindow = aWindow || window;
  window.gui = aWindow.gui = this;

  with (this) {
  this.document = this.window.document;

  try {
    var test = new ActiveXObject( "MSXML2.DOMDocument.3.0" );
  }
  catch (e) {
    return xml3();
  }

  this.xtab = xtab = new XTab( island, this );

  // document state
  this.history = [];
  this.groups = [];
  this.display = "";
  this.sorts = "";

  if ( benchmark )
     document.write( '<textarea id=console cols=80 rows=5></textarea><br>' );
 
  if ( !document.all["contextMenu"] )
    this.menu = initMenus();

  var cmenu = new Menu( "colMenu" );
  cmenu.item( "this", "colClicked", "add or open grouping" );
  cmenu.item();

  for ( var c=0 ; c<xtab.columns.length ; c++ ) {
    var name = xtab.columns[c].name;
    var len;

    if ( !name.match( /^#(\.|$)/ ) )
	this.display += name+",";

    cmenu.item( xtab.columns[c].name = defix(name), "colClicked", "column" );
  }

  cmenu.close( document );

  // determine initial groups, sorts and columns
  this.groups = [(display.match( /=[-+\d$]*[^,]+/g )||[]).concat("").join( "," )
	.replace( /(^|,)=[-+\d$]*/g, "$1" )];
  this.sorts = (display.match( /[-+][^,]+/g )||[]).concat("").join( "," );
  this.display = defix( display ).match( /[^,]+/g ).concat("").join( "," );

  // alert( display+"/"+sorts );
  log( display+"/"+groups+"/"+sorts );

  if ( !window.dataURL )
    aWindow.document.write( "<b id='info'>...</b>" );

  since( "loaded" );
  doUpdate( groups );
  window.onscroll = trackHeader;
  document.all[this.xtab.rname].oncontextmenu = openMenu;
  }

  return this;
}

// user has selected a column op.
function colClicked (item) { with (window.gui) {
  var name = item.innerText;
  var gtype = item.parentElement.parentItem.innerText;

  if ( gtype.match( /Total by/i )|| gtype.match( /Open by/i ) ) {
    groups[groups.length-(groups.length&&
	!groups[groups.length-1].length)?1:0] = name+",";
    if ( gtype.match( /Open by/i ) )
       return openGroup( item );
  }
  else if ( gtype.match( /Group by/i ) ) {
    if ( !groups.length )
      groups[0] = "";

  log( "last group: "+groups[groups.length-1] );
    groups[groups.length-1] =
	groups[groups.length-1].subst( name+"," )+name+",";
  }
  else {
    display = display.subst( "(([^,]+,){"+item.colno+"})", "$1"+name+"," );
    return update();
  }

log( "last group: "+groups[groups.length-1] );

//  if ( !item.level )
//    window.groups = groups;

  xtab.doGroups( groups, item.node );
  update();
} }

// user wants to open up a grouping
function openGroup (item) { with (window.gui) {

  if ( item.innerText == "Close" )
    return closeGroup( item );
  else {
    var gnode, gnodes = item.node.selectNodes( ".//grp" );
    gnodes.length;

    while( gnode = gnodes.nextNode() )
      gnode.setAttribute( "_open", 1 );

    item.node.setAttribute( "_open", 1 );
  }

  update();
} }

// grouping should be closed
function closeGroup (item) {
//alert( item.node.parentNode );
  (item.node.parentNode||item.node).setAttribute( "_open", 0 );
  update();
}

// groupings should be reset
function reGroup (item) { with (window.gui) {
  xtab.unGroup();
  update( groups = [] );
} }

// hide column
function doHide (item) { with (window.gui) {
  display = display.subst( "(([^,]+,){"+item.colno+"})[^,]+,", "$1" );
  update();
} }

// add sort
function doSort (item) { with (window.gui) {
  var cname = item.over.column.name;
  sorts = sorts.subst( "[-+=\d$]"+cname+",", "" )+
     (item.innerText == "Ascending" ? "+" : "-")+cname+",";
  log( "new sort: "+sorts );
  update();
} }

// reset sorts
function unSort (item) { with (window.gui) {
  sorts = "";
} }

// keystroke in cell being edited
function doKey (field) { with (window.gui) {
  var event = window.gui.window.event;
  var over = field.parentElement;

  if ( event.keyCode == 13 ) {
    over.node.setAttribute( over.column.attr, field.value
	.replace( /,/, "" ) );
    update();

    var pKeys = (window.gui).xtab.pKeys[over.column.tab];

    if ( pKeys ) {
        var q = over.column.type.match( /text|date/ ) ? "'" : "";
	var sql = "update "+over.column.tab+
	  " set "+over.column.col+" = "+q+field.value.replace( /'/g, "''" )+q;

	for ( var i=0; i<pKeys.length ; i++ ) {
	  var pkCol = pKeys[i];
          q = pkCol.type.match( /text|date/ ) ? "'" : "";

	  sql += (i==0 ? " where " : " and ")+pkCol.col+
	    " = "+q+(""+over.node.getAttribute( pkCol.attr ))
            .replace( /&amp;#8722;/, "-" ).replace( /&amp;nbsp;/, " " )+q;
	}

	var url = (window.dataURL||window.document.location+"")
			.replace( /\?.*/, "?_sql="+sql );
	document.fetchlet.fetch( url );
	alert( "Sent SQL: "+sql );
    }
  }
  else if ( event.keyCode == 27 ) {
    over.innerHTML = over.saveHTML;
  }

  return true;
} }

// start the edit of a cell
function doEdit (item) {
  var c0 = item.over.parentElement.c0;

  if ( item.over.column.pkey )
	alert( "Can't update key value" );
  else if ( c0 > 0 ) {
    var val = item.over.innerText;
    item.over.saveHTML = item.over.innerHTML;
    item.over.innerHTML = "<input type='textfield' value=\""+
	val+"\" size='"+(val.length+5)+"' onkeypress='return doKey(this)'>";
    item.over.children[0].select();
    item.over.children[0].focus();
  }
  else
    alert( "You can only update \"open\" cells" );
}

// search column for string
function doFind (item) {
  var pattern = prompt( "Find in column "+item.column.name, "" );
  var nodes = item.node.selectNodes( ".//row" );
  var node; nodes.length;

  while ( node = nodes.nextNode() ) {
    if ( node.getAttribute( item.column.attr ).match( pattern ) ) {
	node.setAttribute( "_found", 1 );
	node.parentNode.setAttribute( "_open", 1 );
    }
    else
	node.removeAttribute( "_found" );
  }

  update();
}

// undo last operation
function doUndo (item) { with (window.gui) {

  if ( history.length<2 )
    return alert( "End of history" );

  var last = history[--history.length-1];
  groups  = last[0];
  display = last[1];
  sorts   = last[2];
  xtab.unGroup();
  update( groups );
  history.length--;
} }

function trackHeader () { with (window.gui) {
 if ( !heading.orig ) {       
	heading.orig = heading.parentElement.parentElement.offsetTop;
	heading.style.position = "relative";
 }

 var lockedHeading = parent.frames[0] && parent.frames[0].heading;

 if ( lockedHeading )
    parent.frames[0].document.body.scrollTop = window.document.body.scrollTop

 if ( window.document.body.scrollTop > heading.orig ) {
    heading.style.top = window.document.body.scrollTop-heading.orig-2;

    if (lockedHeading ) {
      lockedHeading.style.position = "relative";
      lockedHeading.style.top  = window.document.body.scrollTop-heading.orig-2;
    }
 }
 else {
    heading.style.top = -1;
    if ( lockedHeading )
      lockedHeading.style.top = -1;
 }
} }

var hcell;

function headerClick () { with (window.gui) {
  var event = window.gui.window.event;

  if ( event.button == 2 )
    return true;

  if ( event.srcElement.tagName == "TH" ) {
    hcell = event.srcElement;
    hcell.leftDown = hcell.offsetLeft;
    hcell.style.position = "relative";
    hcell.xDown = event.clientX;
    hcell.setCapture();
    hwidth = hcell.clientWidth-event.offsetX < 5 ? 22 : null;

    return event.returnValue = false;
  }
  else {
    window.hrow = event.srcElement;
    window.hrow.hdown = event.offsetX;
    window.hrow.cdown = event.clientX;
    return event.returnValue = false;
  }

  return true;
} }

function headerMove () {
  var event = window.gui.window.event;

  if (hcell)
	hcell.style.left = event.clientX - hcell.xDown;

  return event.returnValue = false;
}

function headerDrop () { with (window.gui) {
    var event = window.gui.window.event;

    if ( window.hrow ) {
	var cell, i = 0;

	do
	    cell = window.hrow.children[i++];
	while ( cell.offsetLeft + cell.offsetWidth < window.hrow.hdown )

	var col = xtab.columns[cell.colname];
		col.length = (cell.offsetWidth+
		event.clientX-window.hrow.cdown)/6;

	if ( col.length < 0 )
	   col.length = 0.1;

	window.hrow = null;
	update();
    }

    if ( !hcell )
	return true;

    hcell.releaseCapture();

    var hrow = hcell.parentElement.children;
    var newpos = hcell.leftDown + hcell.clientWidth/2 +
		 event.clientX - hcell.xDown;
    display = "";

    for ( var c=0 ; c<hrow.length ; c++ ) {
	if ( hrow[c] == hcell )
	  continue;

	if ( newpos && hrow[c].offsetLeft + hrow[c].offsetWidth/2 > newpos ) {
	    display += hcell.colname+",";
	    newpos = null;
	}

        display += hrow[c].colname+",";
    }

    if ( newpos && newpos )
      display += hcell.colname+",";

    log( "new order: "+display );
    update();
    hcell = null;
    return event.returnValue = false;
} }

// start update of display
function update ( reGroup ) { with (window.gui) {
  since( "waiting" );
  document.body.style.cursor = "wait";
  setTimeout( reGroup ? "doUpdate( true )" : "doUpdate()", 0 );
} }

// perform update of display
function doUpdate( reGroup ) { with (window.gui) {
  var grps = groups.join("/").subst( ",/(.)", " then $1" );

  if ( document.all["info"] )
	info.innerHTML = (grps ? "Grouping by "+
	grps.substring( 0, grps.length-1 ):"")+
	(sorts ? (grps?" then":"")+" sorting"+
	sorts.substring( 0, sorts.length-1 )
	.replace( /\-/g, " descending by " )
	.replace( /\+/g, " ascending by " ) : "")+
	(grps||sorts?"":" (Right-click on Xtab for menu <a href='/xtab/uguide.html'>user guide</a>)<br>");

  if ( reGroup ) {
    xtab.doGroups( groups );
    since( "grouped " );
  }

  xtab.render( display, sorts );
  since( "written" );

  history[history.length] = [groups.slice(0), display, sorts];
  document.body.style.cursor = "default";

  this.heading = xtab.heading;
  heading.onmousedown = headerClick;
  heading.onmousemove = headerMove;
  heading.onmouseup = headerDrop;
  trackHeader();
} }

// JAVA applet can send requests back to server without a form
if ( 0 ) {
document.write( '<applet codebase="/xtab" code="fetchlet.class"'+
	' name="fetchlet" height=0 width=0 mayscript></applet>' );
fetchlet.style.display = "none";
}

var s = "/script>";

function sidepg () {
 return "<html><link rel='stylesheet' href='/xtab/style.css'>"+
	"<body><div id='pad'></div><div ID=scr></div><br><br><br><br>";
}

function mainpg ( dname, tname ) {
    if ( tname == "undefined" ) tname = "";
    return "<html><body><script>"+
    "document.write( '<link rel=\"stylesheet\" href=\"/xtab/style.css\">' );"+
    (tname&&"document.write( parent.document.all[\""+tname+"\"].xml );")+
    "window.dataURL = 1;\n\n"+
    "window.gui = new parent.Gui( parent[\""+dname+"\"], 0, window );"+
    "window.gui.xtab.render( window.gui.xtab.columns[1].name, "+
	"'', parent.frames[0].scr );"+(tname&&
    "parent.frames[0].pad.style.height = parent.frames[1].pad.offsetHeight;")+
    "top.document.all.ff.cols=(parent.frames[0].scr.offsetWidth)+',*';<"+s;
}

function framepg ( dname, tname ) {
document.write( "<frameset id='ff' border=0 cols='1,*'>"+
"<frame src='javascript: parent.sidepg();' scrolling=no>"+
"<frame src='javascript: parent.mainpg( \""+dname+"\", \""+tname+"\" );'>"+
"</frameset>" );
}
